from distutils.core import setup
import setuptools

setup(
    name = "song_vis",
    version = "1.3.3.2",
    author = "Damith Senanayake",
    author_email = "dracusds123@gmail.com",
    description = "A dimensionality reduction method that retains a parametric model",
    packages = setuptools.find_packages(),
    download_url = 'https://github.com/damithsenanayake/SONG/archive/v1.1.1.tar.gz'
)